/*
 * construir_consultes.c
 * Autor: Sergio Blanco Cuaresma
 * Data: Desembre del 2001
 * Descripci: Funcions per la creaci d'un fitxer verge de hashing
 * i un altre invertit, i per la insercci de dades a cadascun. A
 * ms de la funci de hashing T(...)
 */
#include <stdio.h>
#include "tipus.h"

// Afegeix un mailing al fitxer invertit indicat en l'argument
int afegir_consulta_ma(FILE *fconsulta, unsigned int address, ESTATM estat, DATE data){
  // Contadors per controlar quants mailings hem insertat en cada posici, sn estatics per tant
  // cada cop que es crida la funci encara mantenen el valor de l'ultima crida
  static unsigned int num_pendents=0, num_enviats=0, num_consultats=0, num_afirmatiu=0, num_negatiu=0;
  static unsigned int num_excedents=0;
  struct consulta_mailing consulta;
  struct excedent_mailing excedent;

  consulta.data=data;
  consulta.address=address;
  // Traspasem valor tb a excedent per si la hem d'utilitzar
  excedent.data=data;
  excedent.address=address;
  excedent.estat=estat;
  switch (estat){     // Segons l'estat del mailing l'insertarem en una posici o en un altre
                      // Zona d'excedents sempre a la posici 50000
    case PENDENT:   // Posici 0
      if (num_pendents<10000) {
        fseek(fconsulta, (0+num_pendents)*sizeof(struct consulta_mailing), SEEK_SET);
        if (fwrite(&consulta, sizeof(struct consulta_mailing), 1, fconsulta)!=1){
          return 1;
        }
        num_pendents++;
      } else {  // Zona excedents
        if (num_excedents<15537){
          fseek(fconsulta, num_excedents*sizeof(struct excedent_mailing)+(50000)*sizeof(struct consulta_mailing), SEEK_SET);
          if (fwrite(&excedent, sizeof(struct excedent_mailing), 1, fconsulta)!=1){
            return 1;
          }
          num_excedents++;
        }
      };
      break;

    case ENVIAT:    // Posici 10000
      if (num_enviats<10000) {
        fseek(fconsulta, (10000+num_enviats)*sizeof(struct consulta_mailing), SEEK_SET);
        if (fwrite(&consulta, sizeof(struct consulta_mailing), 1, fconsulta)!=1){
          return 1;
        }
        num_enviats++;
      } else {  // Zona excedents
        if (num_excedents<15537){
          fseek(fconsulta, num_excedents*sizeof(struct excedent_mailing)+(50000)*sizeof(struct consulta_mailing), SEEK_SET);
          if (fwrite(&excedent, sizeof(struct excedent_mailing), 1, fconsulta)!=1){
            return 1;
          }
          num_excedents++;
        }
      };
      break;

    case CONSULTAT:   // Posici 20000
      if (num_consultats<10000) {
        fseek(fconsulta, (20000+num_consultats)*sizeof(struct consulta_mailing), SEEK_SET);
        if (fwrite(&consulta, sizeof(struct consulta_mailing), 1, fconsulta)!=1){
          return 1;
        }
        num_consultats++;
      } else {  // Zona excedents
        if (num_excedents<15537){
          fseek(fconsulta, num_excedents*sizeof(struct excedent_mailing)+(50000)*sizeof(struct consulta_mailing), SEEK_SET);
          if (fwrite(&excedent, sizeof(struct excedent_mailing), 1, fconsulta)!=1){
            return 1;
          }
          num_excedents++;
        }
      };
      break;

    case RESPOSTA_AFIRMATIVA:   // Posici 30000
      if (num_afirmatiu<10000) {
        fseek(fconsulta, (30000+num_afirmatiu)*sizeof(struct consulta_mailing), SEEK_SET);
        if (fwrite(&consulta, sizeof(struct consulta_mailing), 1, fconsulta)!=1){
          return 1;
        }
        num_afirmatiu++;
      } else {  // Zona excedents
        if (num_excedents<15537){
          fseek(fconsulta, num_excedents*sizeof(struct excedent_mailing)+(50000)*sizeof(struct consulta_mailing), SEEK_SET);
          if (fwrite(&excedent, sizeof(struct excedent_mailing), 1, fconsulta)!=1){
            return 1;
          }
          num_excedents++;
        }
      };
      break;

    case RESPOSTA_NEGATIVA:   // Posici 40000
      if (num_negatiu<10000) {
        fseek(fconsulta, (40000+num_negatiu)*sizeof(struct consulta_mailing), SEEK_SET);
        if (fwrite(&consulta, sizeof(struct consulta_mailing), 1, fconsulta)!=1){
          return 1;
        }
        num_negatiu++;
      } else {  // Zona excedents
        if (num_excedents<15537){
          fseek(fconsulta, num_excedents*sizeof(struct excedent_mailing)+(50000)*sizeof(struct consulta_mailing), SEEK_SET);
          if (fwrite(&excedent, sizeof(struct excedent_mailing), 1, fconsulta)!=1){
            return 1;
          }
          num_excedents++;
        }
      };
      break;
  }
  return 0;
};


// Funcio de Hashing, pasem a codi ascii els carcters i els sumem
long T(char *empresa){
  char n=0;
  long num=0;
  
  while ((empresa[n]!='\0') && (n<26)) {
    num += empresa[n];
    n++;
  }
  return num;
}


// Aplica la funci de hashing T() i guarda a la posici que li correspongui que sigui verge.
// "**************************" simbolitza verge.
void escriure_hash(FILE *fhash, char *empresa, unsigned int client_address, long comanda_address){
  long address;
  struct tipo_hash hash;
  int total=0;

  address=T(empresa)*sizeof(struct tipo_hash)*16;	// 16 registres per linea
  if (fseek(fhash, address, SEEK_SET)!=0){ 	// Ens situem a la linea de la zona principal
    perror("Seek fhash");
    exit(1);
  }
  fread(&hash, sizeof(struct tipo_hash), 1, fhash);	// La senya de registre verge sera "****..."
  while ((strncmp(hash.nom, "**************************", 25)!=0) && (total<16) && (!feof(fhash))){
    address += sizeof(struct tipo_hash);
    fread(&hash, sizeof(struct tipo_hash), 1, fhash);
    total++;
  }
  // Si no hi ha espai a la zona principal
  if (strncmp(hash.nom, "**************************", 25)!=0) {		// Busquem i guardem a la zona d'excedents
    address = 3302*sizeof(struct tipo_hash);
    if (fseek(fhash, address, SEEK_SET)!=0){
      perror("Seek fhash");
      exit(1);
    }
    fread(&hash, sizeof(struct tipo_client), 1, fhash);
    total=3302;
    while ((strncmp(hash.nom, "**************************", 25)!=0) && (total<4096) && (!feof(fhash))){
      address += sizeof(struct tipo_hash);
      fread(&hash, sizeof(struct tipo_hash), 1, fhash);
      total++;
    }
  }	// NOTA: D'aquesta condici sempre es surt amb un espai trobat gracies al diseny dels fitxers

  if (fseek(fhash, address, SEEK_SET)!=0){	// Guardem a on hem trobat espai lliure
    perror("Seek fhash");
    exit(1);
  }
  strncpy(hash.nom, empresa, 25);
  hash.client_address=client_address;
  hash.comanda_address=comanda_address;
  if (fwrite(&hash, sizeof(struct tipo_hash), 1, fhash)!=1){
    perror("hashing.dat");
    exit(1);
  }
}


// Crear un fitxer de hashing verge
void init_fitxer_hash(char *fitxer){
  FILE *fhash;
  struct tipo_hash hash;
  unsigned int bucle;
  
  if ((fhash=fopen(fitxer, "wb"))==NULL){
    perror("Init Hashing");
    exit(1);
  }
  strncpy(hash.nom, "*************************\0", 26);   // Simbolitza verge
  for(bucle=0;bucle<4096*16;bucle++){
    fwrite(&hash, sizeof(struct tipo_hash), 1, fhash);
  }
  fclose(fhash);
}


// Crea un fitxer invertit verge
void init_fitxer_ma(char *fitxer){
  struct consulta_mailing consulta;
  struct excedent_mailing excedent;
  unsigned int bucle;
  FILE *fconsulta;

  if ((fconsulta=fopen(fitxer, "wb"))==NULL) { 
    perror ("consulta_a.dat");
    exit(1);
  }
  // Inicialitzaci del fitxer invertit a verge
  // Fitxer invertit Zona Principal
  consulta.data.dia=0;  // Simbolitza verge
  for(bucle=0;bucle<50000;bucle++){
    fwrite(&consulta, sizeof(struct consulta_mailing), 1, fconsulta);
  }
  // Fitxer invertit Zona d'excedents
  excedent.data.dia=0;  // Simbolitza verge
  for(bucle=0;bucle<15536;bucle++){
    fwrite(&consulta, sizeof(struct excedent_mailing), 1, fconsulta);
  }
}
