/*
 * feijen.c
 * Autor: Sergio Blanco Cuaresma
 * Data: Desembre del 2001
 * Descripci: actualitza 3 fitxers mestres (clients, comandes, mailings)
 * a partir d'un fitxer de moviments. A ms aplica funcions de hashing i
 * inserta en un fitxer invertit per poder realitzar certes consultes
 * a posteriori.
 * 
 */
#include <stdio.h>
#include "tipus.h"

FILE *fclients, *fcomandes, *fmailings;
FILE *fclients_a, *fcomandes_a, *fmailings_a;
FILE *fmoviments;
FILE *fhash, *fconsultes_ma;
BOOLEAN final_comandes=FALSE, final_mailings=FALSE;
BOOLEAN no_hi_ha_comandes=FALSE, no_hi_ha_mailings=FALSE;
BOOLEAN no_hi_ha_comandes2=FALSE, no_hi_ha_mailings2=FALSE;   // Variables auxiliars per no perdre valors
struct tipo_client mestre_client, meact_client, client_mov;
struct tipo_comanda *mestre_comanda, *meact_comanda, comanda_mov;
struct tipo_mailing *mestre_mailing, *meact_mailing, mailing_mov;
BOOLEAN kactual_cl, *kactual_co, *kactual_ma;
unsigned int num_comandes, num_mailings; // Nombre de registres agafats del fitxer mestre comandes o mailings
unsigned int num_comandes2, num_mailings2; // Variables auxiliars per no perdre valors
struct seguentreg next;		// Caracterstiques del moviment actual (tipus i acci)
unsigned int mov_k;		// Codi_client del moviment actual
unsigned int k;

// Funcions externes per fer el hashing i crear el fitxer invertit
extern int afegir_consulta_ma(FILE *fconsulta, unsigned int address, ESTATM estat, DATE data);
extern void escriure_hash(FILE *fhash, char *empresa, unsigned int client_address, long comanda_address);
extern void init_fitxer_hash(char *fitxer);
extern void init_fitxer_ma(char *fitxer);


// Obrim tots el fitxers que necessitarem al llarg de l'execuci
void obrir_arxius() {
  if ((fclients=fopen("clients.dat", "rb"))==NULL) {
    perror ("clients.dat");
    exit(1);
  }
  if ((fcomandes=fopen("comandes.dat", "rb"))==NULL) {
    perror ("comandes.dat");
    exit(1);
  }
  if ((fmailings=fopen("mailings.dat", "rb"))==NULL) {
    perror ("mailings.dat");
    exit(1);
  }
  if ((fmoviments=fopen("moviments.dat", "rb"))==NULL) {
    perror ("moviments.dat");
    exit(1);
  }
  if ((fclients_a=fopen("clients_a.dat", "wb"))==NULL) {
    perror ("clients_a.dat");
    exit(1);
  }
  if ((fcomandes_a=fopen("comandes_a.dat", "wb"))==NULL) {
    perror ("comandes_a.dat");
    exit(1);
  }
  if ((fmailings_a=fopen("mailings_a.dat", "wb"))==NULL) {
    perror ("arxiu mailings_a.dat");
    exit(1);
  }
  // Inicialitzem l'arxiu invertit de mailings a verge
  init_fitxer_ma("consultes_a.dat");
  if ((fconsultes_ma=fopen("consultes_a.dat", "rb+"))==NULL) {
    perror ("arxiu consultes_a.dat");
    exit(1);
  }
  // Inicialitzem l'arxiu de hashing a verge
  init_fitxer_hash("hashing_a.dat");
  if ((fhash=fopen("hashing_a.dat", "rb+"))==NULL) {
    perror ("arxiu hashing_a.dat");
    exit(1);
  }
}


// Tanquem la resta d'arxius que quedaran oberts desprs d'aplicar l'algorisme i
// actualitzarem els fitxers mestres amb els nous.
void tancar_arxius(){
  fclose(fclients_a);
  fclose(fcomandes_a);
  fclose(fmailings_a);
  fclose(fconsultes_ma);
  fclose(fhash);
  
  // El arxius mestres actualitzats pasen a ser el arxius mestres finals
  if (unlink("clients.dat")!=0){
    perror("Borrar clients.dat");
    exit(1);
  }
  if (rename("clients_a.dat", "clients.dat")!=0){
    perror("Renombrar clients_a.dat");
    exit(1);
  }
  if (unlink("comandes.dat")!=0){
    perror("Borrar comandes.dat");
    exit(1);
  }
  if (rename("comandes_a.dat", "comandes.dat")!=0){
    perror("Renombrar comandes_a.dat");
    exit(1);
  }
  if (unlink("mailings.dat")!=0){
    perror("Borrar mailings.dat");
    exit(1);
  }
  if (rename("mailings_a.dat", "mailings.dat")!=0){
    perror("Renombrar mailings_a.dat");
    exit(1);
  }
  if (unlink("consultes.dat")!=0){
    perror("Borrar consultes.dat");
    exit(1);
  }
  if (rename("consultes_a.dat", "consultes.dat")!=0){
    perror("Renombrar consultes_a.dat");
    exit(1);
  }
  if (unlink("hashing.dat")!=0){
    perror("Borrar hashing.dat");
    exit(1);
  }
  if (rename("hashing_a.dat", "hashing.dat")!=0){
    perror("Renombrar hashing_a.dat");
    exit(1);
  }
}


// Llegim un client del fitxer mestre, a ms, llegirem les seves comandes i
// mailings si existeixen (s'indicara la existencia d'aqueste amb variables booleanes).
void lectura_mestre(void) {
  // Llegim client
  fread(&mestre_client, sizeof(struct tipo_client), 1, fclients);
  if (feof(fclients)) {
    mestre_client.codi_client = 65535;        // Infinit
    close(fclients);
  } else {
    if (final_comandes == FALSE){   // Si no s'acabat el fitxer de comandes
      // Solicitem memria
      if ((mestre_comanda = (struct tipo_comanda*)malloc(sizeof(struct tipo_comanda)))==NULL) {
        perror("Comanda memria");
        exit(1);
      }
      num_comandes=0;		// Inicialitzem el nmero de comandes llegides
      fread(&mestre_comanda[num_comandes], sizeof(struct tipo_comanda), 1, fcomandes);
      // Llegim totes les comandes que pertanyin a aquest client
      while ((mestre_comanda[num_comandes].codi_client == mestre_client.codi_client) && (!feof(fcomandes))){
        num_comandes++;
        if ((mestre_comanda = (struct tipo_comanda*)realloc(mestre_comanda, ((num_comandes+1)*sizeof(struct tipo_comanda))))==NULL){
          perror("Comanda memria");
          exit(1);
        }
        fread(&mestre_comanda[num_comandes], sizeof(struct tipo_comanda), 1, fcomandes);
      }
      if ((feof(fcomandes)) && (num_comandes==0)){	// Final del fitxer
        free(mestre_comanda);
        fclose(fcomandes);
        no_hi_ha_comandes= final_comandes= TRUE;
      } else {
        if (num_comandes==0) {		// No hi ha cap comanda d'aquest client
          free(mestre_comanda);
          no_hi_ha_comandes=TRUE;
        } else {  // L'ultim registre llegit no ens interesa (pertany al seguen client)
          no_hi_ha_comandes=FALSE;
          num_comandes--;
          if ((mestre_comanda = (struct tipo_comanda*)realloc(mestre_comanda, ((num_comandes+1)*sizeof(struct tipo_comanda))))==NULL){
            perror("Comanda del memria");
            exit(1);
          }
        }
        // Apuntem un altre cop en l'arxiu, al registre que hem eliminat de memria
        fseek(fcomandes, (-1)*sizeof(struct tipo_comanda), SEEK_CUR);
      }
    }
    if (final_mailings == FALSE){   // Si no s'ha acabat el fitxer de mailings
      // Solicitem memria
      if ((mestre_mailing = (struct tipo_mailing*)malloc(sizeof(struct tipo_mailing)))==NULL) {
        perror("Mailing memria");
        exit(1);
      }
      num_mailings=0;		// Inicialitzem el nmero de mailings llegits
      fread(&mestre_mailing[num_mailings], sizeof(struct tipo_mailing), 1, fmailings);
      // Llegim tots els mailings que pertanyen a aquest client
      while ((mestre_mailing[num_mailings].codi_client == mestre_client.codi_client) && (!feof(fmailings))){
        num_mailings++;
        if ((mestre_mailing = (struct tipo_mailing*)realloc(mestre_mailing, ((num_mailings+1)*sizeof(struct tipo_mailing))))==NULL){
          perror("Mailing memria");
          exit(1);
        }
        fread(&mestre_mailing[num_mailings], sizeof(struct tipo_mailing), 1, fmailings);
      }
      if ((feof(fmailings)) && (num_mailings==0)){	// Final del fitxer
        free(mestre_mailing);
        fclose(fmailings);
        no_hi_ha_mailings = final_mailings = TRUE;
      } else { 
        if (num_mailings==0){   // No hi ha cap mailing d'aquest client
          free(mestre_mailing);
          no_hi_ha_mailings = TRUE;
        } else {  // L'ultim registre llegit no ens interesa
          no_hi_ha_mailings = FALSE;
          num_mailings--;
          if ((mestre_mailing = (struct tipo_mailing*)realloc(mestre_mailing, ((num_mailings+1)*sizeof(struct tipo_mailing))))==NULL){
            perror("Mailing del memria");
            exit(1);
          }
        } 
      }
      // Apuntem un altre cop en l'arxiu, al registre que hem eliminat de memria
      fseek(fmailings, (-1)*sizeof(struct tipo_mailing), SEEK_CUR);
    }
  }
}


// Retorna el nmero mnim de 2 introduits com parmetres
unsigned int minim (unsigned int a, unsigned int b) {
  if (a < b) {
    return a;
  } else {
    return b;
  }
}


// Llegim moviment, primer es fa una lectura del tipus de moviment que ve a continuacio
// i depenen de quin es tracti, es llegeix un client o un mailing o una comanda. A ms
// fem que mov_k contingui el codi de client del moviment llegit.
void lectura_moviment() {
  // Llegim registre que ens indicara de quin tipus es el moviment (i la seva accio [alta, baixa...])
  fread(&next, sizeof(struct seguentreg), 1, fmoviments);
  if (feof(fmoviments)) {
    mov_k = 65535;    // Infinit
    fclose(fmoviments);
    return;
  }

  if (next.tipus == CLIENT) {
    fread(&client_mov, sizeof(struct tipo_client), 1, fmoviments);
    mov_k = client_mov.codi_client;
  } else if (next.tipus == COMANDA) {
    fread(&comanda_mov, sizeof(struct tipo_comanda), 1, fmoviments);
    mov_k = comanda_mov.codi_client;
  } else if (next.tipus == MAILING) {
    fread(&mailing_mov, sizeof(struct tipo_mailing), 1, fmoviments);
    mov_k = mailing_mov.codi_client;
  }
}


// Si tenim a memria el client al qual volem aplicar el moviment o un ms petit
// creem un nou espai de memoria (meact..) sobre el que farem les modificiacions
// junt amb el seu corresponent kactual, tb salvem diversos valors que els necesitarem
// mes endavant. A ms es fa una nova lectura sobre el fitxer mestre.
void obtenir_estat_ini_clau(void){
  unsigned int bucle;

  if (k==mestre_client.codi_client){
    meact_client=mestre_client;
    kactual_cl=TRUE;
    if ((final_comandes==FALSE) && (no_hi_ha_comandes==FALSE)) {
      // Solicitem espai per mestre actualitzat
      if ((meact_comanda = (struct tipo_comanda*)malloc((num_comandes+1)*sizeof(struct tipo_comanda)))==NULL){
        perror("Comandes act memria");
        exit(1);
      }
      // Solicitem espai per kactual
      if ((kactual_co = (BOOLEAN *)malloc((num_comandes+1)*sizeof(BOOLEAN)))==NULL) {
        perror("Kactual Comandes memria");
        exit(1);
      }
      for (bucle=0; bucle<=num_comandes; bucle++){
        meact_comanda[bucle]=mestre_comanda[bucle];
        kactual_co[bucle]=TRUE;
      }
    }
    if ((final_mailings==FALSE) && (no_hi_ha_mailings==FALSE)){
      // Solicitem espai per mestre actualitzat
      if ((meact_mailing = (struct tipo_mailing*)malloc((num_mailings+1)*sizeof(struct tipo_mailing)))==NULL){
          perror("Mailings act memria");
          exit(1);
      }
      // Solicitem espai per kactual
      if ((kactual_ma = (BOOLEAN *)malloc((num_mailings+1)*sizeof(BOOLEAN)))==NULL) {
        perror("Kactual Mailings memria");
        exit(1);
      }
      for (bucle=0; bucle<=num_mailings; bucle++){
        meact_mailing[bucle]=mestre_mailing[bucle];
        kactual_ma[bucle]=TRUE;
      }
    }
    no_hi_ha_comandes2=no_hi_ha_comandes;
    no_hi_ha_mailings2=no_hi_ha_mailings;
    num_comandes2=num_comandes;
    num_mailings2=num_mailings;
    lectura_mestre();
  } else {
    no_hi_ha_comandes = no_hi_ha_comandes2 = TRUE;
    no_hi_ha_mailings = no_hi_ha_mailings2 = TRUE;
    kactual_cl=FALSE;
  }
}


// Realitzem moviment sobre el client actual
void actualitzar_estat_clau_client(){
  switch (next.accio){
    case ALTA:
      if (kactual_cl==TRUE) {
        fprintf(stderr, "ANOMALIA: Estem intentant donar d'alta un client que ja existeix\n");
      } else {
        meact_client=client_mov;
        kactual_cl=TRUE;
      }
      break;
    case BAIXA:
      if (kactual_cl==TRUE){
        kactual_cl=FALSE;
      } else {
        fprintf(stderr, "ANOMALIA: Estem intentant donar de baixa un client que no existeix\n");
      }
      break;
    case MODIFICACIO:
      if (kactual_cl==TRUE){
        meact_client=client_mov;
      } else {
        fprintf(stderr, "ANOMALIA: Estem intentan modificar un client que no existeix\n");
      }
      break;
  }
}


// Busca a *meact_comandes[] l'element que te el mateix codi_comanda que comanda_mov
// Retorna una estructura indicant si ha sigut trobat i en quina posici
struct X trobar_mateix_codi_comanda(){
  struct X X;
  unsigned int bucle;

  X.no_trobat=TRUE;
  if (no_hi_ha_comandes2==FALSE) {    // Noms busca si t comandes
    for(bucle=0;bucle<=num_comandes2;bucle++){
      if (meact_comanda[bucle].codi_comanda==comanda_mov.codi_comanda){
        X.no_trobat=FALSE;
        break;
      }
    }
  }
  X.pos=bucle;
  return X;
}

// Actualitzem dades mestres a memoria quan els moviments son de tipus comanda
void actualitzar_estat_clau_comanda(){ 
  struct X X;

  if (kactual_cl==FALSE){   // Si el client no existeix, no hi poden haver comandes d'ell
    switch (next.accio) {
      case ALTA:
        fprintf(stderr, "ANOMALIA: Estem intentant donar d'alta una comanda d'un client que no existeix\n");
        break;
      case BAIXA:
        fprintf(stderr, "ANOMALIA: Estem intentant donar de baixa una comanda d'un client que no existeix\n");
        break;
      case MODIFICACIO:
        fprintf(stderr, "ANOMALIA: Estem intentant modificar una comanda d'un client que no existeix\n");
        break;
    }
    return; // Fi
  }
  
  X=trobar_mateix_codi_comanda();   // Ens retorna la posici on es troba el registre mestre a modificar
  switch (next.accio) {
    case ALTA:
      if ((X.no_trobat==FALSE) && (kactual_co[X.pos]==TRUE)) {
        fprintf(stderr, "ANOMALIA: Estem intentant donar d'alta una comanda que ja existeix\n");
      } else {
        if (no_hi_ha_comandes2==TRUE) {   // Creem l'espai de memria meact que no s'havia creat abans pq
                                          // aquest client no tenia comandes
          if ((meact_comanda=(struct tipo_comanda*)malloc((num_comandes2+1)*sizeof(struct tipo_comanda)))==NULL){
            perror("Alta comanda memria");
            exit(1);
          }
          if ((kactual_co=(BOOLEAN*)malloc((num_comandes2+1)*sizeof(BOOLEAN)))==NULL) {
            perror("Alta comanda kactual memria");
            exit(1);
          }
          num_comandes2=0;
          no_hi_ha_comandes2=FALSE;

        } else {    // Ampliem memria i afegim al final
          num_comandes2++;
            if ((meact_comanda=(struct tipo_comanda*)realloc(meact_comanda, (num_comandes2+1)*sizeof(struct tipo_comanda)))==NULL){
              perror("Alta comanda memria");
              exit(1);
            }
            if ((kactual_co=(BOOLEAN*)realloc(kactual_co, (num_comandes2+1)*sizeof(BOOLEAN)))==NULL) {
              perror("Alta comanda kactual memria");
              exit(1);
            }
        }
        meact_comanda[num_comandes2]=comanda_mov;   // Afegim el nou registre
        kactual_co[num_comandes2]=TRUE;
      }
      break;

    case BAIXA:
      if ((X.no_trobat==FALSE) && (kactual_co[X.pos]==TRUE)) {
        kactual_co[X.pos]=FALSE;
      } else {
        fprintf(stderr, "ANOMALIA: Estem intentant donar de baixa una comanda que no existeix\n");
      }
      break;
    case MODIFICACIO:
      if ((X.no_trobat==FALSE) && (kactual_co[X.pos]==TRUE)) {
        meact_comanda[X.pos]=comanda_mov;
      } else {
        fprintf(stderr, "ANOMALIA: Estem intentant modificar una comanda que no existeix\n");
      }
      break;
  }
}
	

// Busca a *meact_mailings[] l'element que te el mateix codi_mailing que mailing_mov
// Retorna una estructura indicant si ha sigut trobat i en quina posici
struct X trobar_mateix_codi_mailing(){
  struct X X;
  unsigned int bucle;

  X.no_trobat=TRUE;
  if (no_hi_ha_mailings2==FALSE) {    // Noms busca si t mailings
    for(bucle=0;bucle<=num_mailings2;bucle++){
      if (meact_mailing[bucle].codi_mailing==mailing_mov.codi_mailing){
        X.no_trobat=FALSE;
        break;
      }
    }
  }
  X.pos=bucle;
  return X;
}


// Actualitzem dades mestres a memoria quan els moviments son de tipus mailing
void actualitzar_estat_clau_mailing(){
  struct X X;

  if (kactual_cl==FALSE){   // Si no existeix el client, no hi poden haver mailings
    switch (next.accio) {
      case ALTA:
        fprintf(stderr, "ANOMALIA: Estem intentant donar d'alta un mailing d'un client que no existeix\n");
        break;
      case BAIXA:
        fprintf(stderr, "ANOMALIA: Estem intentant donar de baixa un mailing d'un client que no existeix\n");
        break;
      case MODIFICACIO:
        fprintf(stderr, "ANOMALIA: Estem intentant modificar un mailing d'un client que no existeix\n");
        break;
    }
    return;
  }

  X=trobar_mateix_codi_mailing();   // Ens retorna la posici on es troba el registre mestre a modificar
  switch (next.accio) {
    case ALTA:
      if ((X.no_trobat==FALSE) && (kactual_ma[X.pos]==TRUE)) {
          fprintf(stderr, "ANOMALIA: Estem intentant donar d'alta una mailing que ja existeix\n");
      } else {
        if (no_hi_ha_mailings2==TRUE) { // Creem l'espai de memria meact que no s'havia creat abans pq
                                        // aquest client no tenia mailings
          if ((meact_mailing=(struct tipo_mailing*)malloc((num_mailings2+1)*sizeof(struct tipo_mailing)))==NULL){
            perror("Alta mailing memria");
            exit(1);
          }
          if ((kactual_ma=(BOOLEAN*)malloc((num_mailings2+1)*sizeof(BOOLEAN)))==NULL) {
            perror("Alta mailing kactual memria");
            exit(1);
          }
          num_mailings2=0;
          no_hi_ha_mailings2=FALSE;

        } else {  // Ampliem memria i afegim al final
          num_mailings2++;
          if ((meact_mailing=(struct tipo_mailing*)realloc(meact_mailing, (num_mailings2+1)*sizeof(struct tipo_mailing)))==NULL){
            perror("Alta mailing memria");
            exit(1);
          }
          if ((kactual_ma=(BOOLEAN*)realloc(kactual_ma, (num_mailings2+1)*sizeof(BOOLEAN)))==NULL) {
            perror("Alta mailing kactual memria");
            exit(1);
          }
        }
        meact_mailing[num_mailings2]=mailing_mov;   // Afegim el nou registre
        kactual_ma[num_mailings2]=TRUE;
      }
      break;
      
    case BAIXA:
      if ((X.no_trobat==FALSE) && (kactual_ma[X.pos]==TRUE)) {
        kactual_ma[X.pos]=FALSE;
      } else {
        fprintf(stderr, "ANOMALIA: Estem intentant donar de baixa una mailing que no existeix\n");
      }
      break;
    case MODIFICACIO:
      if ((X.no_trobat==FALSE) && (kactual_ma[X.pos]==TRUE)) {
        meact_mailing[X.pos]=mailing_mov;
      } else {
        fprintf(stderr, "ANOMALIA: Estem intentant modificar una mailing que no existeix\n");
      }
      break;
  }
}


// Guardem el client i les seves comandes i mailings que acabem de tractar. A ms apliquem
// la funci de hashing i afegim al fitxer invertit (noms mailings).
void guardar(void){
  unsigned int bucle;

  // Guardem client
  if ((fwrite(&meact_client, sizeof(struct tipo_client), 1, fclients_a))!=1){
    perror("Escriptura clients actualitzat");
    exit(1);
  }

  // Hashing per la consulta per nom d'empresa
  if (no_hi_ha_comandes2==FALSE) {
    escriure_hash(fhash, meact_client.nom, (ftell(fclients_a)/sizeof(struct tipo_client))-1, ftell(fcomandes_a));
  } else {
    // Quan el camp d'adrea de comandes cont un 1, simbolitza que no hi han. La posici 1 mai sera utilitzada
    // ja que sempre comena des del 0 i el proxim valor seria 0+sizeof(struct tipo_comanda) i no 1.
    escriure_hash(fhash, meact_client.nom, (ftell(fclients_a)/sizeof(struct tipo_client))-1, 1);
  }
  
  if (no_hi_ha_comandes2==FALSE) {    // Si hi ha comandes
    for(bucle=0;bucle<=num_comandes2;bucle++) {
      if (kactual_co[bucle]==TRUE) {    // Guardem les indicades
        if ((fwrite(&meact_comanda[bucle], sizeof(struct tipo_comanda), 1, fcomandes_a))!=1){
          perror("Escriptura comandes actualitzat");
          exit(1);
        }
      }
    }
    free(kactual_co);   // Alliberem memria
    free(meact_comanda);
  }

  if (no_hi_ha_mailings2==FALSE) {  // Si hi ha mailings
    for(bucle=0;bucle<=num_mailings2;bucle++) {
      if (kactual_ma[bucle]==TRUE) {  // Guardem els indicats
        if ((fwrite(&meact_mailing[bucle], sizeof(struct tipo_mailing), 1, fmailings_a))!=1){
	        perror("Escriptura mailing actualitzat");
	        exit(1);
	      }
        // Afegim al fitxer invertit
        // Mailing a la posici: (ftell(fmailings_a)/sizeof(struct tipo_mailing))-1
        // Hem de resta 1 pq ara ja apuntem a la seguent posici i ens interesa la anterior.
        if (afegir_consulta_ma(fconsultes_ma, (ftell(fmailings_a)/sizeof(struct tipo_mailing))-1, meact_mailing[bucle].estat, meact_mailing[bucle].data)!=0){
          perror("Afegir consulta");
          exit(1);
        }
      }
    }
    free(kactual_ma);   // Alliberem memria
    free(meact_mailing);
  }
}


// Principal
int main(void) {
  obrir_arxius();
  lectura_mestre();
  lectura_moviment();
  k = minim(mestre_client.codi_client, mov_k);
  while (k!=65535){
    obtenir_estat_ini_clau();
    // Processar tots el moviments
    while (k==mov_k){
      switch(next.tipus) {
        case CLIENT:
          actualitzar_estat_clau_client();
          break;
        case COMANDA:
          actualitzar_estat_clau_comanda();
          break;
        case MAILING:
          actualitzar_estat_clau_mailing();
          break;
      }
      lectura_moviment();
    }
    if (kactual_cl==TRUE){
      guardar();
    }
    k = minim(mestre_client.codi_client, mov_k);
  }
  tancar_arxius();
}
